/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of mst_rdb
 *
 * mst_rdb is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * mst_rdb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

/* rdb_col_read_st.c */

#include "mst_rdb.h"

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>
#include <suplib/io.h>

/*

  PROCEDURE
    rdb_col_read_st - Grab columns of data from an /rdb data file.

  DESCRIPTION
    The rdb_col_read_st function extracts data from an /rdb file on
    a stream.  It reads the next record on the stream, extracting data via
    the specified column mapping and fills in a user supplied structure.

    An NULL string is returned if an RDB_String column contains no data, the
    user must still free the pointer to the NULL string. 
    An empty RDB_Num column is flagged as an error if it is supposed to
    contain data. Empty lines are ignored.

    When reading string arguments, this routine places into the target
    data structure a pointer to a dynamically allocated string area, into
    which the string is copied.  When the user is finished using the
    string, this area should be freed.

  RETURNS
    The routine returns 1 if it successfully read data, 0 upon end
    of data. If the number of columns in the record is not the same as in
    the header, or if a (numeric) column contains garbage, a message is
    printed to stderr and the program is halted.

  AUTHOR
    Diab Jerius, Richard J. Edgar

*/
int
rdb_col_read_st(
  FILE *fin,			/* the stream to parse */
  rdbHeader *hdr,		/* the rdb file description */
  DataColumnMap_st const *map,	/* the columns to read */
  void * data			/* where to stow the data. must be pointer
                                   to an appropriate struct */
)
{
  unsigned long i;
  char *buf,  *p0, *p1, *localdata;
  unsigned long j = 0;

  tf_enter("rdb_col_read_st");

  localdata = (char *)data;
  /* reject empty lines */
  while ( (buf = fget_rec(fin, hdr->buf)) )
  {
    hdr->line++;
    if (buf[0]) break;
  }

  if (! buf)
  {
    tf_leave();
    return 0;
  }

  /* initialize pointers to record */
  p0 = p1 = buf;

  for (i = 0 ; i < map->n ; i++)
  {
    char *p, *p2, *pend;

    for (; j < map->map[i].column && p1 ; j++)
    {
      p1 = strchr(p0, '\t');
      p0 = p1 + 1;
    }

    /* check if prematurely reached end of record */
    if (! p1)
      tf_exit(ExitERR_misc, "not enough columns at line %d", hdr->line);

    /* seal off end of column.  if there are no more tabs, it's the last one
       in the record */
    if ( (pend = strchr(p0, '\t')) ) *pend = '\0';

    /* get data and process possible error conditions */
    errno = 0;
    if(map->map[i].type == RDB_Num)
    { * (double *)(localdata + map->map[i].offset) 
           = strtod(p0, &p);
    /* check for garbage after data */
    for (p2 = p ; *p2 && isspace(*p2) ; p2++);

    /* strtod on SunOS is non-ANSI and doesn't set errno to ERANGE if the
       string is empty;  thus the second condition in the following
       statement */
       if (errno || p == p0 || (p[0] && *p2))
       {
         char *name = hdr->col_name[map->map[i].column];

         /* check if empty column */
         if (p == p0 && (!p[0] || (isspace(p[0]) && ! *p2)) ) {
	   if ( map->map[i].type == RDB_String ) {
	     char *str = malloc( 1 );
	     if (! str )
	       tf_exit(ExitERR_alloc,
		       "unable to allocate space for RDB_String");
	     *( char **) ( localdata + map->map[i].offset ) = str;
	     *str = '\0';
	   } else
	     tf_exit(ExitERR_misc,
		     "no value in column %s(%d) at line %d", name, j+1, hdr->line
);
	 }

         /* check if there's garbage in the column */
         if (p[0] && *p2)
           tf_exit(ExitERR_misc,
                    "illegal value in column %s(%d) at line %d: `%s'",
                      name, j+1, hdr->line, p0);

         /* must be overflow or underflow. underflow results in a warning, not
         an error */
         if (errno == ERANGE && (double)*(localdata+map->map[i].offset) == 0)
           tf_message("numerical underflow in column %s(%d) at line %d: `%s'\n",
                            name, j+1, hdr->line, p0);
         else
           tf_exit(ExitERR_misc,
                        "numerical overflow in column %s(%d) at line %d: `%s'",
                                      name, j+1, hdr->line, p0);
       }
    }
    else  /* assume RDB_String if not RDB_Num */
    {  
	char *str = malloc(strlen(p0)+1);
	if (! str )
	  tf_exit(ExitERR_alloc,
		"unable to allocate space for RDB_String");
	*( char **) ( localdata + map->map[i].offset ) = str;
	strcpy(str, p0);
    }

    /* reset EOS to tab */
    if (pend) *pend = '\t';
  }

  for (; j < hdr->n && p1 ; j++)
  {
    p1 = strchr(p0, '\t');
    p0 = p1 + 1;
  }

  /* check if the rest of the record is legitimate, complain if there
     are fewer columns than in the header.  if there are more, we can't
     get to 'em anyway.  this also avoids the problem of having trailing
     tabs which shouldn't be there, but which are harmless. */
  if (j < hdr->n)
    tf_exit(ExitERR_misc,
      "inconsistent number of columns at line %d", hdr->line);

  tf_leave();
  return 1;
}

